# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
from __future__ import annotations

import datetime
import os
import re

import ydb

from airflow import DAG
from airflow.decorators import task
from airflow.operators.python import PythonOperator
from airflow.providers.ydb.hooks.ydb import YDBHook
from airflow.providers.ydb.operators.ydb import YDBExecuteQueryOperator

# [START ydb_operator_howto_guide]


# create_pet_table, populate_pet_table, get_all_pets, and get_birth_date are examples of tasks created by
# instantiating the YDB Operator

ENV_ID = os.environ.get("SYSTEM_TESTS_ENV_ID")
DAG_ID = "ydb_operator_dag"


@task
def populate_pet_table_via_bulk_upsert():
    hook = YDBHook()
    column_types = (
        ydb.BulkUpsertColumns()
        .add_column("pet_id", ydb.OptionalType(ydb.PrimitiveType.Int32))
        .add_column("name", ydb.PrimitiveType.Utf8)
        .add_column("pet_type", ydb.PrimitiveType.Utf8)
        .add_column("birth_date", ydb.PrimitiveType.Utf8)
        .add_column("owner", ydb.PrimitiveType.Utf8)
    )

    rows = [
        {"pet_id": 3, "name": "Lester", "pet_type": "Hamster", "birth_date": "2020-06-23", "owner": "Lily"},
        {"pet_id": 4, "name": "Quincy", "pet_type": "Parrot", "birth_date": "2013-08-11", "owner": "Anne"},
    ]
    hook.bulk_upsert("pet", rows=rows, column_types=column_types)


def sanitize_date(value: str) -> str:
    """Ensure the value is a valid date format"""
    if not re.fullmatch(r"\d{4}-\d{2}-\d{2}", value):
        raise ValueError(f"Invalid date format: {value}")
    return value


def transform_dates(**kwargs):
    begin_date = sanitize_date(kwargs.get("begin_date"))
    end_date = sanitize_date(kwargs.get("end_date"))
    return {"begin_date": begin_date, "end_date": end_date}


with DAG(
    dag_id=DAG_ID,
    start_date=datetime.datetime(2020, 2, 2),
    schedule="@once",
    catchup=False,
) as dag:
    # [START ydb_operator_howto_guide_create_pet_table]
    create_pet_table = YDBExecuteQueryOperator(
        task_id="create_pet_table",
        sql="""
            CREATE TABLE pet (
            pet_id INT,
            name TEXT NOT NULL,
            pet_type TEXT NOT NULL,
            birth_date TEXT NOT NULL,
            owner TEXT NOT NULL,
            PRIMARY KEY (pet_id)
            );
          """,
        is_ddl=True,  # must be specified for DDL queries
    )

    # [END ydb_operator_howto_guide_create_pet_table]
    # [START ydb_operator_howto_guide_populate_pet_table]
    populate_pet_table = YDBExecuteQueryOperator(
        task_id="populate_pet_table",
        sql="""
              UPSERT INTO pet (pet_id, name, pet_type, birth_date, owner)
              VALUES (1, 'Max', 'Dog', '2018-07-05', 'Jane');

              UPSERT INTO pet (pet_id, name, pet_type, birth_date, owner)
              VALUES (2, 'Susie', 'Cat', '2019-05-01', 'Phil');
            """,
    )
    # [END ydb_operator_howto_guide_populate_pet_table]
    # [START ydb_operator_howto_guide_get_all_pets]
    get_all_pets = YDBExecuteQueryOperator(task_id="get_all_pets", sql="SELECT * FROM pet;")
    # [END ydb_operator_howto_guide_get_all_pets]
    transform_dates = PythonOperator(
        task_id="transform_dates",
        python_callable=transform_dates,
        op_kwargs={"begin_date": "{{params.begin_date}}", "end_date": "{{params.end_date}}"},
        params={"begin_date": "2020-01-01", "end_date": "2020-12-31"},
    )
    # [START ydb_operator_howto_guide_get_birth_date]
    get_birth_date = YDBExecuteQueryOperator(
        task_id="get_birth_date",
        sql="""
        SELECT * FROM pet WHERE birth_date BETWEEN '{{ ti.xcom_pull(task_ids="transform_dates")["begin_date"] }}' AND '{{ ti.xcom_pull(task_ids="transform_dates")["end_date"] }}'
        """,
    )
    # [END ydb_operator_howto_guide_get_birth_date]

    (
        create_pet_table
        >> populate_pet_table
        >> populate_pet_table_via_bulk_upsert()
        >> get_all_pets
        >> transform_dates
        >> get_birth_date
    )
    # [END ydb_operator_howto_guide]

    from tests_common.test_utils.watcher import watcher

    # This test needs watcher in order to properly mark success/failure
    # when "tearDown" task with trigger rule is part of the DAG
    list(dag.tasks) >> watcher()

from tests_common.test_utils.system_tests import get_test_run  # noqa: E402

# Needed to run the example DAG with pytest (see: tests/system/README.md#run_via_pytest)
test_run = get_test_run(dag)



# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
from __future__ import annotations

from unittest.mock import patch

from airflow.models.connection import Connection
from airflow.providers.ydb.utils.credentials import get_credentials_from_connection

TEST_ENDPOINT = "my_endpoint"
TEST_DATABASE = "/my_db"
MAGIC_CONST = 42


@patch("ydb.StaticCredentials")
def test_static_creds(mock):
    mock.return_value = MAGIC_CONST
    c = Connection(conn_type="ydb", host="localhost", login="my_login", password="my_pwd")
    credentials = get_credentials_from_connection(TEST_ENDPOINT, TEST_DATABASE, c, {})
    assert credentials == MAGIC_CONST

    assert len(mock.call_args.args) == 1
    driver_config = mock.call_args.args[0]
    assert driver_config.endpoint == TEST_ENDPOINT
    assert driver_config.database == TEST_DATABASE
    assert mock.call_args.kwargs == {"user": "my_login", "password": "my_pwd"}


@patch("ydb.AccessTokenCredentials")
def test_token_creds(mock):
    mock.return_value = MAGIC_CONST
    c = Connection(conn_type="ydb", host="localhost")
    credentials = get_credentials_from_connection(TEST_ENDPOINT, TEST_DATABASE, c, {"token": "my_token"})
    assert credentials == MAGIC_CONST

    mock.assert_called_with("my_token")


@patch("ydb.AnonymousCredentials")
def test_anonymous_creds(mock):
    mock.return_value = MAGIC_CONST
    c = Connection(conn_type="ydb", host="localhost")
    credentials = get_credentials_from_connection(TEST_ENDPOINT, TEST_DATABASE, c)
    assert credentials == MAGIC_CONST
    mock.assert_called_once()


@patch("ydb.iam.auth.MetadataUrlCredentials")
def test_vm_metadata_creds(mock):
    mock.return_value = MAGIC_CONST
    c = Connection(conn_type="ydb", host="localhost")
    credentials = get_credentials_from_connection(TEST_ENDPOINT, TEST_DATABASE, c, {"use_vm_metadata": True})
    assert credentials == MAGIC_CONST
    mock.assert_called_once()


@patch("ydb.iam.auth.BaseJWTCredentials.from_content")
def test_service_account_json_creds(mock):
    mock.return_value = MAGIC_CONST
    c = Connection(conn_type="ydb", host="localhost")

    credentials = get_credentials_from_connection(
        TEST_ENDPOINT, TEST_DATABASE, c, {"service_account_json": "my_json"}
    )
    assert credentials == MAGIC_CONST
    mock.assert_called_once()

    assert mock.call_args.args == ("my_json",)


@patch("ydb.iam.auth.BaseJWTCredentials.from_file")
def test_service_account_json_path_creds(mock):
    mock.return_value = MAGIC_CONST
    c = Connection(conn_type="ydb", host="localhost")

    credentials = get_credentials_from_connection(
        TEST_ENDPOINT, TEST_DATABASE, c, {"service_account_json_path": "my_path"}
    )
    assert credentials == MAGIC_CONST
    mock.assert_called_once()

    assert mock.call_args.args == ("my_path",)


def test_creds_priority():
    # 1. static creds
    with patch("ydb.StaticCredentials") as mock:
        c = Connection(conn_type="ydb", host="localhost", login="my_login", password="my_pwd")
        mock.return_value = MAGIC_CONST
        credentials = get_credentials_from_connection(
            TEST_ENDPOINT,
            TEST_DATABASE,
            c,
            {
                "service_account_json": "my_json",
                "service_account_json_path": "my_path",
                "use_vm_metadata": True,
                "token": "my_token",
            },
        )
        assert credentials == MAGIC_CONST
        mock.assert_called_once()

    # 2. token
    with patch("ydb.AccessTokenCredentials") as mock:
        c = Connection(conn_type="ydb", host="localhost", password="my_pwd")
        mock.return_value = MAGIC_CONST
        credentials = get_credentials_from_connection(
            TEST_ENDPOINT,
            TEST_DATABASE,
            c,
            {
                "service_account_json": "my_json",
                "service_account_json_path": "my_path",
                "use_vm_metadata": True,
                "token": "my_token",
            },
        )
        assert credentials == MAGIC_CONST
        mock.assert_called_once()

    # 3. service account json path
    with patch("ydb.iam.auth.BaseJWTCredentials.from_file") as mock:
        c = Connection(conn_type="ydb", host="localhost")
        mock.return_value = MAGIC_CONST
        credentials = get_credentials_from_connection(
            TEST_ENDPOINT,
            TEST_DATABASE,
            c,
            {
                "service_account_json": "my_json",
                "service_account_json_path": "my_path",
                "use_vm_metadata": True,
            },
        )
        assert credentials == MAGIC_CONST
        mock.assert_called_once()

    # 4. service account json
    with patch("ydb.iam.auth.BaseJWTCredentials.from_content") as mock:
        c = Connection(conn_type="ydb", host="localhost")
        mock.return_value = MAGIC_CONST
        credentials = get_credentials_from_connection(
            TEST_ENDPOINT,
            TEST_DATABASE,
            c,
            {
                "service_account_json": "my_json",
                "use_vm_metadata": True,
            },
        )
        assert credentials == MAGIC_CONST
        mock.assert_called_once()

    # 5. vm metadata
    with patch("ydb.iam.auth.MetadataUrlCredentials") as mock:
        c = Connection(conn_type="ydb", host="localhost")
        mock.return_value = MAGIC_CONST
        credentials = get_credentials_from_connection(
            TEST_ENDPOINT,
            TEST_DATABASE,
            c,
            {
                "use_vm_metadata": True,
            },
        )
        assert credentials == MAGIC_CONST
        mock.assert_called_once()

    # 6. anonymous
    with patch("ydb.AnonymousCredentials") as mock:
        c = Connection(conn_type="ydb", host="localhost")
        mock.return_value = MAGIC_CONST
        credentials = get_credentials_from_connection(TEST_ENDPOINT, TEST_DATABASE, c, {})
        assert credentials == MAGIC_CONST
        mock.assert_called_once()
